# Pfilter/Generate.pl
#
# library functions to generate output commands

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# functions in this file:

# function to convert an interface name to it's broadcast address
# sub generate_interface_to_broadcast( $vars_ref, $interface, 
# 				       $error_lines_ref, $warning_lines_ref )

# function to convert an address or address/mask to a list of
# sub generate_range_to_interfaces( $vars_ref, $range, 
# 				    $error_lines_ref, $warning_lines_ref )

# function to output a comment block showing the configuratoin file
# sub generate_show_config_line( $vars_ref, $source_lines_ref, $directive_ref, 
# 			         $error_lines_ref, $warning_lines_ref )n

# function to generate all of the intermediate source output lines.
# sub generate_source( $vars_ref, $source_lines_ref,
# 		       $error_lines_ref, $warning_lines_ref )

# function to generate the source code for a single ALIAS directive
# sub generate_source_alias_directive( $vars_ref, $directive_ref,
# 				       $pseudo_interface_index,
# 				       $source_lines_ref,
# 				       $error_lines_ref,
# 				       $warning_lines_ref )

# function to generate the source code for a single FORWARD directive
# sub generate_source_forward_directive( $vars_ref, $directive_ref,
# 				         $source_lines_ref,
# 				         $error_lines_ref,
# 				         $warning_lines_ref )

# function to generate any required forwarding/masquerading source code
# sub generate_source_forwarding_masquerading( $vars_ref,
# 					       $source_lines_ref,
# 					       $error_lines_ref,
# 					       $warning_lines_ref )

# function to generate the source code for a single literal block
# sub generate_source_literal_block( $vars_ref, $directive_ref,
# 				     $source_lines_ref,
# 				     $error_lines_ref,
# 				     $warning_lines_ref )

# function to generate the source code for a single open/close directive
# sub generate_source_open_close_directive( $vars_ref, 
# 					    $directive_ref,
# 					    $source_lines_ref,
# 					    $error_lines_ref,
# 					    $warning_lines_ref )

# function to generate the open/close source lines for a directive
# sub generate_source_open_close_lines( $vars_ref, $directive_ref,
# 				        $opening_flag, 
# 				        $source_interfaces_ref,
# 				        $destinations_ref, 
# 				        $source_lines_ref,
# 				        $error_lines_ref, 
# 				        $warning_lines_ref )

# function to generate code for open/close/alias/literal/forward directives
# sub generate_source_standard_directives( $vars_ref, $source_lines_ref,
# 					   $error_lines_ref, $warning_lines_ref )

# function to generate the source code for a syntax error directive
# sub generate_source_syntax_error_directive( $vars_ref, $directive_ref,
# 					      $source_lines_ref,
# 					      $error_lines_ref,
# 					      $warning_lines_ref )

# function to generate the source code for the portions of a
# sub generate_source_trust_address_directive( $vars_ref, $directive_ref,
# 					       $source_lines_ref,
# 					       $error_lines_ref,
# 					       $warning_lines_ref )

#=======================================================================

# function to convert an interface name to it's broadcast address
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   interface name or "ANY"
#   reference to error lines array
#   reference to warning lines array

sub generate_interface_to_broadcast {
    my ( $vars_ref, $interface, $error_lines_ref, $warning_lines_ref ) = @_;

#    print "$0: entering generate_interface_to_broadcast interface=$interface\n"
#	if ${$$vars_ref{constants}}{debug};

    my $broadcast = "UNKNOWN";

    # if not passed "ANY" as the interface name, get it's broadcast address

    my $interfaces_ref = $$vars_ref{interfaces};
    if ( exists $$interfaces_ref{$interface} ) {
	my $interface_ref = $$interfaces_ref{$interface};
	if ( exists $$interface_ref{broadcast} ) {
	    $broadcast = $$interface_ref{broadcast};
	}
    }
	 
    print "$0: exiting  generate_interface_to_broadcast interface=$interface broadcast=$broadcast\n"
	if ${$$vars_ref{constants}}{debug};

    return $broadcast;
}

#=======================================================================

# function to convert an address or address/mask to a list of
# interface names that can include the address or address/mask
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   address or address/mask string
#   reference to error lines array
#   reference to warning lines array

sub generate_range_to_interfaces {
    my ( $vars_ref, $range, $error_lines_ref, $warning_lines_ref ) = @_;
    print "$0: entering generate_range_to_interfaces range=$range\n"
	if ${$$vars_ref{constants}}{debug};
    my @interfaces = ();

    # if not passed "ANY" the address/range, figure out the interfaces list
    @interfaces = range_to_interfaces( $vars_ref,
				       $range,
				       $error_lines_ref,
				       $warning_lines_ref )
        if ( $range ne "ANY" );

    # if we were passed "ANY" as the address/range, or we couldn't
    # figure out a list of interfaces for the passed address/range,
    # return a single interface name of "ANY"
    @interfaces = ( "ANY" )
        if ! @interfaces;

    if ( ${$$vars_ref{constants}}{debug} ) {
        my $string = join( ',', @interfaces );
        print "$0: exiting  generate_range_to_interfaces range=$range returning ($string)\n";
    }
    return @interfaces;
}

#=======================================================================

# function to output a comment block showing the configuratoin file
# source line that caused generated output
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to source lines list to add generated lines to
#   reference to directive in question
#   reference to error lines array
#   reference to warning lines array

sub generate_show_config_line {
    my ( $vars_ref, $source_lines_ref, $directive_ref, 
	 $error_lines_ref, $warning_lines_ref ) = @_;
    print "$0: entering generate_show_config_line\n"
	if ${$$vars_ref{constants}}{debug};

    push @$source_lines_ref, 
    "  # Configuration file line $$directive_ref{config_line_number}:";
    push @$source_lines_ref, "  #";
    push @$source_lines_ref, "  # \t$$directive_ref{config_line}";
    push @$source_lines_ref, "  #";
    push @$source_lines_ref, "  # generated this output:";
    push @$source_lines_ref, "  #";
}

#=======================================================================

# function to generate all of the intermediate source output lines.
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source {
    my ( $vars_ref, $source_lines_ref, $error_lines_ref, 
	 $warning_lines_ref ) = @_;
    print "$0: entering generate_source\n"
	if ${$$vars_ref{constants}}{debug};

    # this macro is always at the start of the intermediate file
    push @$source_lines_ref, "%pfilter_prefix%"
        if ! ${$$vars_ref{constants}}{noprefix};
    
    # generate all normal output caused by the straightforward directives
    generate_source_standard_directives( $vars_ref,
                                         $source_lines_ref,
                                         $error_lines_ref,
                                         $warning_lines_ref );

    # generate any forwarding/masquerading stuff between interfaces
    generate_source_forwarding_masquerading( $vars_ref,
                                             $source_lines_ref,
                                             $error_lines_ref,
                                             $warning_lines_ref );
    
    # this macro is always at the end of the intermediate file
    push @$source_lines_ref, "%pfilter_suffix%"
        if ! ${$$vars_ref{constants}}{nosuffix};
    
    # check for configuration file directives that weren't used when
    # generating the intermediate source file
    foreach my $directive_ref ( @{$$vars_ref{config}} ) { # loop through directives
        if ( ! $$directive_ref{processed} ) {          # if directive not processed ...
            my $directive_type = $$directive_ref{directive_type};
            if ( $directive_type ne "comment_lines" # if directive is a non-comment and ...
                 && $directive_type ne "logging"    # it is not a LOGGING which generates nada ..
                 && $directive_type ne "nologging"  # it is not a NOLOGGING which generates nada ..
                 && $directive_type ne "nomerge"    # it is not a NOMERGE which generates nada ..
                 && $directive_type ne "verbose"    # it is not VERBOSE or DEBUG which are ...
                 && $directive_type ne "debug" ) {  # processed before building structures
                print "$0: internal error - generating output did not use configuration file line $$directive_ref{config_line_number}:\n";
                print "$0:           $$directive_ref{config_line}\n";
            } 
        }
    }
    print "$0: exiting  generate_source\n" 
        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================

# function to generate the source code for a single ALIAS directive
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to directive being used to generate code
#   index number of pseudo interface to be created
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_alias_directive {
    my ( $vars_ref,
	 $directive_ref,
	 $pseudo_interface_index,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;
    print "$0: entering generate_source_alias_directive\n"
	if ${$$vars_ref{constants}}{debug};

    # speed things up
    my $faked_address = $$directive_ref{faked_address};
    my $real_address = $$directive_ref{real_address};

    # mark this directive as being processed
    $$directive_ref{processed} = 1;
    print_hash("", "generating output from directive", $directive_ref)
        if ${$$vars_ref{constants}}{debug};

    # put a line in the output showing the original configuration
    # file line that caused the following output to be generated
    generate_show_config_line( $vars_ref,
                               $source_lines_ref,
                               $directive_ref,
                               $error_lines_ref,
                               $warning_lines_ref );
    # point to the destination protocols/ports hash
    my $protocols_ports_ref = 
        $$directive_ref{destination_protocols_ports};

    # point to the ruleset defined network services to open
    my $services_ref = $$directive_ref{destination_services};

    # if the ALIAS directive included protocols/ports to open, 
    # or included ruleset defined network services to open ...
    if ( keys %$protocols_ports_ref || @$services_ref ) {

        # generate the open/close output lines
        my @interfaces = ( "ANY" );
        my @destinations = ( $real_address );
        generate_source_open_close_lines( $vars_ref,
					  $directive_ref,
					  1,
					  \@interfaces,
					  \@destinations,
					  $source_lines_ref,
					  $error_lines_ref,
					  $warning_lines_ref );

#        push @$source_lines_ref, "  #";

    } # end if the ALIAS specified protocols/ports to open

    # find out which interface the faked address will be created on
    # and figure out the netmask and broadcast address of that interface

    my $interface_name = address_to_interface( $vars_ref,
					       $faked_address,
					       $error_lines_ref,
					       $warning_lines_ref );
    my $interfaces_ref = $$vars_ref{interfaces};
    my $interface_ref = $$interfaces_ref{$interface_name};
    my $interface_mask = $$interface_ref{mask};
    my $interface_broadcast = $$interface_ref{broadcast};
    print_hash("", "in generate_source_alias_directive interface", $interface_ref)
        if ${$$vars_ref{constants}}{debug};

    # figure out the numeric ip addresses in case iptables needs them

    my $faked_ip = address_to_ip( $vars_ref,
				  $faked_address,
				  $error_lines_ref,
				  $warning_lines_ref );
    my $real_ip = address_to_ip( $vars_ref,
				 $real_address,
				 $error_lines_ref,
				 $warning_lines_ref );

    # put a line in the output to invoke the alias setup macro

    push @$source_lines_ref, 
    "  %pfilter_alias% $faked_address $faked_ip $real_address $real_ip $interface_name $pseudo_interface_index $interface_mask $interface_broadcast";
    
    print "$0: exiting  generate_source_alias_directive\n" 
        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================

# function to generate the source code for a single FORWARD directive
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to directive being used to generate code
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_forward_directive {
    my ( $vars_ref,
	 $directive_ref,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;
    print "$0: entering generate_source_forward_directive\n"
	if ${$$vars_ref{constants}}{debug};

    # mark this directive as being processed
    $$directive_ref{processed} = 1;
    print_hash("", "generating output from directive", $directive_ref)
        if ${$$vars_ref{constants}}{debug};

    # put a line in the output showing the original configuration
    # file line that caused the following output to be generated
    generate_show_config_line( $vars_ref,
                               $source_lines_ref,
                               $directive_ref,
                               $error_lines_ref,
                               $warning_lines_ref );

    # figure out the destination addresses, it's either
    # specified as a list of specific address(ranges), or 
    # no destination addresses were specified which means
    # that the destination is this computer in which case
    # we set the destinations to all of our interface addresses

    my @destinations;
    if ( exists $$directive_ref{destination_addresses} &&
	 scalar @{$$directive_ref{destination_addresses}} ) {
        @destinations = @{$$directive_ref{destination_addresses}};
    } else {
	@destinations = @{$$vars_ref{interface_addresses}};
    }

    # Figure out the sources to which this applies, ANY
    # means any source addresses, the macro will take care
    # of including or not including the source flags.

    my @sources;
    if ( exists $$directive_ref{source_addresses} &&
	 @{$$directive_ref{source_addresses}} ) {
        @sources = @{$$directive_ref{source_addresses}};
    } else {
        @sources = ( "ANY" );
    }

    # convert the onto address into an absolute ip address in case
    # we are in some mode like iptables where some things require
    # an absolute ip address rather than a dns address

    my $onto_address_ip = address_to_ip( $vars_ref,
                                         $$directive_ref{onto_address},
                                         $error_lines_ref,
				         $warning_lines_ref );

    # find the forwarded protocols/ports and services
    my $protocols_ports_ref = 
        $$directive_ref{forwarded_protocols_ports};
    my @protocols = sort keys %$protocols_ports_ref;
    my $services_ref = $$directive_ref{forwarded_services};

    # for each forwarded protocol listed
    # in the protocols/ports hash, ...

    foreach my $protocol ( @protocols ) {
			    
        # for each port listed in the sub-hash
			    
        my $ports_ref = $$protocols_ports_ref{$protocol};
        my @ports_list = keys %$ports_ref;
        foreach my $port_index ( 0 .. $#ports_list ) {
	    my $port = $ports_list[$port_index];
				    
	    # if there were parsed forwarded onto from
	    # protocols/ports in the directive, we have
	    # to match the particular port index for
	    # that protocol, otherwise we just use the 
	    # same port that is being forwarded from
	    
	    my $onto_port = $port;
	    if ( exists $$directive_ref{onto_protocols_ports} ) {
		my $onto_protocols_ports_ref = 
		    $$directive_ref{onto_protocols_ports};
		if ( ! exists $$onto_protocols_ports_ref{$protocol} ) {
		    print "$0: internal error - generating output found forwarded protocol $protocol with onto protocols but without onto protocol $protocol for configuration file line $$directive_ref{config_line_number}:\n";	
		} else {
		    my $onto_ports_ref = 
			$$onto_protocols_ports_ref{$protocol};
		    my @onto_ports_list = keys %$onto_ports_ref;
		    if ( $port_index > $#onto_ports_list ) {
			print "$0: internal error - generating output forwarded protocol $protocol port index $port_index port value $port with onto protocols/ports without port at same index for configuration file line $$directive_ref{config_line_number}:\n";	
		    } else {
			$onto_port = $onto_ports_list[$port_index];
		    }
		}
	    }
	    
	    # for each original packet destination address, ...
	    
	    foreach my $destination ( @destinations ) {
#                print "$0: in       generate_source_forward_directive destination=<$destination>\n"
#	            if ${$$vars_ref{constants}}{debug};
		
		# if packet source interfaces were specified for
		# this directive, skip this destination address
		# if it can't come in on this interface

		if ( exists $$directive_ref{interfaces} &&
		     @{$$directive_ref{interfaces}} ) {
		    print "$0: in       generate_source_forward_directive checking destination $destination match against specified interfaces <@{$$directive_ref{interfaces}}>\n"
			if ${$$vars_ref{constants}}{debug};
		    my @possible_interfaces = 
			generate_range_to_interfaces( $vars_ref,
						      $destination,
						      $error_lines_ref,
						      $warning_lines_ref );
		    my @intersected_interfaces =
			interfaces_intersection( $vars_ref,
						 \@possible_interfaces,
						 $$directive_ref{interfaces} );
	            if ( ! @intersected_interfaces ) {
			print "$0: in       generate_source_forward_directive skipping destination  $destination specified=<@{$$directive_ref{interfaces}}> possible=<@possible_interfaces>\n"
			    if ${$$vars_ref{constants}}{debug};
	                next;
	            }    
	        }

		# if this destination is not one of this machine's 
		# interface addresses, find a list of the possible
		# interface names that could be used to output
		# packets to this destination
		
		my @destination_interfaces =
		    grep( /^$destination$/, @{$$vars_ref{interface_addresses}} ) ?
			( "ANY" ) :
			    generate_range_to_interfaces
				( $vars_ref,
				  $destination,
				  $error_lines_ref,
				  $warning_lines_ref );
		
		# for each source address, ...
		
		foreach my $source ( @sources ) {
#                print "$0: in       generate_source_forward_directive source=<$source>\n"
#	            if ${$$vars_ref{constants}}{debug};
		    
		    # find the possible source interfaces that
		    # this source address could come from

		    my @source_interfaces =
			generate_range_to_interfaces
			    ( $vars_ref,
			      $source,
			      $error_lines_ref,
			      $warning_lines_ref );

		    # if packet source interfaces were specified,
		    # intersect that list with the possible 
		    # source interfaces that were just found,
		    # and if that results on no common interfaces,
		    # skip this source address
		    
		    if ( exists $$directive_ref{interfaces} &&
			 @{$$directive_ref{interfaces}} ) {
			print "$0: in        generate_source_forward_directive source_interfaces=<@source_interfaces> directive_interfaces=<@{$$directive_ref{interfaces}}>\n"
			    if ${$$vars_ref{constants}}{debug};
			@source_interfaces =
			    interfaces_intersection( $vars_ref,
						     \@source_interfaces,
						     $$directive_ref{interfaces} );
			print "$0: in        generate_source_forward_directive new source_interfaces=<@source_interfaces>\n"
			    if ${$$vars_ref{constants}}{debug};
		        next if ! @source_interfaces;
		    }

		    # if this source/destination is appropriate,
		    # or if this destination is "ANY",
		    # or path optimizations are disabled, ...
		    
		    if ( ${$$vars_ref{constants}}{nooptpath} ||
			$destination eq "ANY" ||
			    valid_source_destination_combination 
				( $vars_ref,
				  $source, 
				  $destination, 
				  $error_lines_ref,
				  $warning_lines_ref ) ) {
		    
		        # one little tricky thing is that if the 
		        # destination address is the faked address
		        # in an aliased host, the real address 
		        # needs to be used instead for the ipchains/
		        # iptables/whatever generated rule because
		        # filtering is done in the kernel after
		        # translating the destination address,
		        # this also means re-calculating the
		        # list of possible destination interfaces,
		        # and we also want to output some comment
		        # explanation lines to explain this if
		        # we haven't already output a comment
		        # block explaing this for this destination.
		        # Of course, this is meaningless if the
		        # destination is "ANY".
		    
		        my $alias_faked_to_real_ref = 
			    $$vars_ref{alias_faked_to_real};
			if ( $destination ne "ANY" &&
			     exists $$alias_faked_to_real_ref{$destination} ) {
			    
			    my $real_destination = 
				$$alias_faked_to_real_ref{$destination};
			    @destination_interfaces =
				generate_range_to_interfaces
				    ( $vars_ref,
				      $real_destination,
				      $error_lines_ref,
				      $warning_lines_ref );
			    push @$source_lines_ref,
			    "  # Since $destination is the pseudo address of an ALIAS,";
			    push @$source_lines_ref,
			    "  # we use the real address $real_destination because that is what the";
			    push @$source_lines_ref,
			    "  # packet matching will see after the packet has been remapped.";
			    $using_alias_real_notified_flag = 1;
			    $destination = $real_destination;
			}
		    
			# for every source interface, ...
			
			foreach my $source_interface ( @source_interfaces ) {
			
			    # for every destination interface, ...
			    
			    foreach my $destination_interface ( @destination_interfaces ) {
				
				push @$source_lines_ref,
				"  %pfilter_forward_protocol_port% $protocol $port $source $destination $source_interface $destination_interface $$directive_ref{onto_address} $onto_address_ip $onto_port";
				
				
			    } # end for each destination interface
		
			} # end for each source interface
	    
		    } # end if the source/destination is appropriate

	        } # end of loop for each source address

	    } # end of loop for each destination address

        } # end of loop for each port listed

    } # end of loop for each protocol listed

    print "$0: exiting  generate_source_forward_directive\n" 
        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================

# function to generate any required forwarding/masquerading source code
# (not to be confused with generated code for the FORWARD directive)
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_forwarding_masquerading {
    my ( $vars_ref,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};
    print "$0: entering generate_source_forwarding_masquerading\n"
	if $debug;

    # if we have both protected and unprotected interfaces, do the ip
    # masquerading of the packets going out on unprotected interfaces.
    # the reason this code block is here instead of being done in the
    # pfilter-prefix or pfilter-suffix macros, is that the command
    # lines being inserted into the output file include both the 
    # interface name and the interface ip address on the same line 
    # and there is no easy way to do this in the rulesets.

    my $interfaces_ref = $$vars_ref{interfaces};
    if ( ${$$vars_ref{constants}}{unprotected_interface_names} ne ""
        && ${$$vars_ref{constants}}{protected_interface_names} ne "" ) {

        # if there were any configuration lines setting interfaces
        # as protected/unprotected, add them to the output commands
        # as comments saying which lines generated this
        my $config_lines = 0;
        foreach my $directive_ref ( @{$$vars_ref{config}} ) {
            my $directive_type = $$directive_ref{directive_type};
            if ( $directive_type eq "extif"       ||
                 $directive_type eq "filtered"    ||
                 $directive_type eq "intif"       ||
                 $directive_type eq "protected"   ||
                 $directive_type eq "trusted"     ||
                 $directive_type eq "unfiltered"  ||
                 $directive_type eq "unprotected" ||
                 $directive_type eq "untrusted" ) {
		if ( ! $config_lines ) {
		    push @$source_lines_ref, 
		    "  # These configuration file line(s):";
		    push @$source_lines_ref, "  #";
		    $config_lines = 1;
		}
                push @$source_lines_ref,
                "  # $$directive_ref{config_line_number}\t$$directive_ref{config_line}";
            }
        }
        if ( $config_lines ) {
	    push @$source_lines_ref, "  #";
            push @$source_lines_ref, "  # generated this output:";
        }
        push @$source_lines_ref, "  #" if $config_lines;

        # push out comment lines explaining this

        push @$source_lines_ref,
        "  # Since we have both protected and unprotected interfaces,";
        push @$source_lines_ref,
        "  # that means we need to configure ip masquerading (otherwise";
        push @$source_lines_ref,
        "  # known as NAT for Network Address Translation), and enable";
        push @$source_lines_ref,
        "  # appropriate packet forwarding between interfaces.";
        push @$source_lines_ref, "  #";

        # for each interface that is marked as unprotected
        # and that we know the external address for ...

        foreach my $outside_interface_name ( keys %$interfaces_ref ) {
            print "$0: in       generate_source_forwarding_masquerading possible outside_interface_name=$outside_interface_name\n"
                if $debug;
            my $outside_interface_ref = 
            $$interfaces_ref{$outside_interface_name};
            print_hash( "", "in       generate_source_forwarding_masquerading outside_interface",
                         $outside_interface_ref ) if $debug;
            if ( ! $$outside_interface_ref{protected} 
                 && exists $$outside_interface_ref{address} ) {
                print "$0: in       generate_source_forwarding_masquerading outside_interface_name=$outside_interface_name\n"
                    if $debug;

		# for each interface that is marked as protected ...

		foreach my $inside_interface_name ( keys %$interfaces_ref ) {
                    print "$0: in       generate_source_forwarding_masquerading possible inside_interface_name=$inside_interface_name\n"
                        if $debug;
		    my $inside_interface_ref = 
			$$interfaces_ref{$inside_interface_name};
		    if ( $$inside_interface_ref{protected}  ) {
                        print "$0: in       generate_source_forwarding_masquerading inside_interface_name=$inside_interface_name\n"
                            if $debug;

			# Do whatever this flavor of packet filtering
                        # requires, to make packets coming into this
                        # system from the inside (protected) interface
			# that are to be forwarded out to the outside 
                        # (unprotected) interface, look like they 
                        # originated from the interface address on the 
                        # outside (unprotected) interface. In other
                        # words, enable/force NATTING (network address
                        # translation) of connections originating from
                        # the inside (protected) interface that are 
			# going out to the outside (unprotected)
                        # interface.

                        my $inside_interface_address_mask = 
                            ( $$inside_interface_ref{mask} && 
                            $$inside_interface_ref{mask} ne "" ) ?
			    "$$inside_interface_ref{address}/$$inside_interface_ref{mask}" :
			    $$inside_interface_ref{address};
                        push @$source_lines_ref, 
                        "  %pfilter_nat% $inside_interface_name $inside_interface_address_mask $outside_interface_name $$outside_interface_ref{address}";

                        # We have to decide whether to add 
                        # output rules to allow the outgoing 
                        # forwarded packets. We have to be careful 
                        # to not let them be forwarded directly
                        # into this machine's external address,
                        # if the protected interface is an
                        # untrusted/filtered interface

                        print_hash( "       ", 
                                    "in       generate_source_forwarding_masquerading inside_interface",
                                    $inside_interface_ref )
                            if $debug;
                        if ( $$inside_interface_ref{filtered} ) {
                            push @$source_lines_ref,
                            "  # ";
                            push @$source_lines_ref,
                            "  # Since the protected interface $$inside_interface_ref{name} ($inside_interface_address_mask)";
                            push @$source_lines_ref,
                            "  # is marked as being filtered (untrusted), we don't want to";
                            push @$source_lines_ref,
                            "  # allow connections directly to this machine from the protected";
                            push @$source_lines_ref,
                            "  # interface $$inside_interface_ref{name} ($inside_interface_address_mask).  But since these";
                            push @$source_lines_ref,
                            "  # rules are being put in both INPUT and the FORWARD chains, and";
                            push @$source_lines_ref,
                            "  # we are about to allow ALL packets to be forwarded from the";
                            push @$source_lines_ref,
                            "  # protected interface $$inside_interface_ref{name} to the unprotected interface $$outside_interface_ref{name}";
                            push @$source_lines_ref,
                            "  # as part of the ip masquerading, which will also allow those";
                            push @$source_lines_ref,
                            "  # packets destined for this machine, we first need to add a";
                            push @$source_lines_ref,
                            "  # a specific rule to block packets from being forwarded from";
                            push @$source_lines_ref,
                            "  # the protected interface $$inside_interface_ref{name} to this machine's $$outside_interface_ref{name} address";
                            push @$source_lines_ref,
                            "  # $$outside_interface_ref{address}.  This probably isn't needed, since the following";
                            push @$source_lines_ref,
                            "  # code that allows outgoing connections to be forwarded should";
                            push @$source_lines_ref,
                            "  # not apply to packets going into this machine, but better to";
                            push @$source_lines_ref,
                            "  # be safe than sorry.";
                            push @$source_lines_ref,
                            "  #";
                            push @$source_lines_ref,
                            "  %close_protocol_port% ANY $$outside_interface_ref{address} $$inside_interface_ref{name} ANY ANY ANY";
                        }
                        push @$source_lines_ref,
                        "  # ";
                        push @$source_lines_ref,
                        "  # Allow outgoing connections to be forwarded from the";
                        push @$source_lines_ref,
                        "  # protected interface $$inside_interface_ref{name} to the unprotected interface $$outside_interface_ref{name}.";
                        push @$source_lines_ref,
                        "  #";
                        push @$source_lines_ref,
                        "  %open_protocol_port% ANY ANY $$inside_interface_ref{name} $$outside_interface_ref{name} ANY ANY";

		    }
		}
            }
        }
        push @$source_lines_ref, "  #";
        push @$source_lines_ref, "  # Enable packet forwarding between interfaces.";
        push @$source_lines_ref, "  #";
        push @$source_lines_ref, "  echo 1 > /proc/sys/net/ipv4/ip_forward";
        push @$source_lines_ref, "";
    }

    print "$0: exiting  generate_source_forwarding_masquerading\n" 
        if $debug;
}

#=======================================================================

# function to generate the source code for a single literal block
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to directive being used to generate code
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_literal_block {
    my ( $vars_ref,
	 $directive_ref,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;
#    print "$0: entering generate_source_literal_block\n"
#	if ${$$vars_ref{constants}}{debug};

    # mark this directive as being processed
    $$directive_ref{processed} = 1;

    # put a line in the output showing what this is
    push @$source_lines_ref,
    "  # $$directive_ref{literal_config_line_number}\t$$directive_ref{literal_config_line}";
    push @$source_lines_ref, "  #";

    # put the literal lines into the output
    push @$source_lines_ref, @{$$directive_ref{literal_lines_ref}};
    
    # put some lines in the output ending the block
    push @$source_lines_ref, "  #";
    push @$source_lines_ref,
    "  # $$directive_ref{endliteral_config_line_number}\t$$directive_ref{endliteral_config_line}";
    
#    print "$0: exiting  generate_source_literal_block\n" 
#        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================

# function to generate the source code for a single open/close directive
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to directive being used to generate code
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_open_close_directive {
    my ( $vars_ref,
	 $directive_ref,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;

#    print "$0: entering generate_source_open_close_directive\n"
#	if ${$$vars_ref{constants}}{debug};

    # mark this directive as being processed

    $$directive_ref{processed} = 1;
    print_hash("", "generating output from directive", $directive_ref)
        if ${$$vars_ref{constants}}{debug};

    # put a line in the output showing the original configuration
    # file line that caused the following output to be generated

    generate_show_config_line( $vars_ref,
                               $source_lines_ref,
                               $directive_ref,
                               $error_lines_ref,
                               $warning_lines_ref );

    # figure out the destination addresses, it's either
    # specified as a list of specific address(ranges), or 
    # no destination addresses were specified which means
    # that the destination is this computer in which case
    # we set the destinations to ANY if we are in
    # --nooptpath mode, or to all the filtered interface
    # addresses since the unfiltered interfaces accept
    # everything going to this machine

    my @destinations;
    if ( exists $$directive_ref{destination_addresses} &&
	 scalar @{$$directive_ref{destination_addresses}} ) {
        @destinations = @{$$directive_ref{destination_addresses}};
    } else {
	@destinations = ( ${$$vars_ref{constants}}{nooptpath} ) ?
	    ( "ANY" ) : @{$$vars_ref{interface_addresses}};
    }

    # figure out the possible source interfaces, it's either
    # specified as a list of specific interfaces, or no
    # interfaces were specified in which case we use "ALL"

    my @interfaces;
    if ( exists $$directive_ref{interfaces} &&
	 @{$$directive_ref{interfaces}} ) {
        @interfaces = @{$$directive_ref{interfaces}};
    } else {
        @interfaces = ( "ANY" );
    }

    # generate the open/close output lines

    my $opening_flag = ( $$directive_ref{directive_type} eq "open" ) ?
                         1 : 0;
    generate_source_open_close_lines( $vars_ref,
				      $directive_ref,
				      $opening_flag,
                                      \@interfaces,
				      \@destinations,
				      $source_lines_ref,
				      $error_lines_ref,
				      $warning_lines_ref );

    print "$0: exiting  generate_source_open_close_directive\n" 
        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================

# function to generate the open/close source lines for a directive
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to directive being used to generate code
#   a flag, non-zero if opening, zero if closing
#   reference to list of source interfaces
#   reference to list of destinations
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_open_close_lines {
    my ( $vars_ref,
	 $directive_ref,
	 $opening_flag,
	 $source_interfaces_ref,
	 $destinations_ref,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;

    if ( ${$$vars_ref{constants}}{debug} ) {
        my $source_interfaces = join( ',', @$source_interfaces_ref );
        my $destinations = join( ',', @$destinations_ref );
	print "$0: entering generate_source_open_close_lines opening=$opening_flag source_interfaces=($source_interfaces) destinations=($destinations)\n"
    }

    # pick the macro to invoke the opens/closes,
    # and pick the ruleset defined service action

    my $close_open_macro;
    my $ruleset_action;
    if ( $opening_flag ) {
	$close_open_macro = "%open_protocol_port%";
	$ruleset_action = "open";
    } else {
	$close_open_macro = "%close_protocol_port%";
	$ruleset_action = "close";
    }

    # Figure out the sources to which this applies, ANY
    # means any source addresses, the macro will take care
    # of including or not including the source flags.

    my @sources;
    if ( exists $$directive_ref{source_addresses} &&
	 @{$$directive_ref{source_addresses}} ) {
        @sources = @{$$directive_ref{source_addresses}};
    } else {
        @sources = ( "ANY" );
    }

    # for each destination address, ...

    foreach my $destination ( @$destinations_ref ) {

        # clear the flag that says we have written output
	# comments about translating faked to real alias
	# address for this destination

	my $using_alias_real_notified_flag = 0;
	
	# if this destination is not one of this machine's 
	# interface addresses, find a list of the possible
	# interface names that could be used to output
	# packets to this destination

	my @destination_interfaces =
	    grep( /^$destination$/, @{$$vars_ref{interface_addresses}} ) ?
		( "ANY" ) :
		generate_range_to_interfaces
		    ( $vars_ref,
		      $destination,
		      $error_lines_ref,
		      $warning_lines_ref );
	    
	# for each source address, ...

	foreach my $source ( @sources ) {

	    # find the possible source interfaces that
	    # this source address (range) could come from
	    
	    my @source_address_interfaces =
		generate_range_to_interfaces
		    ( $vars_ref,
		      $source,
		      $error_lines_ref,
		      $warning_lines_ref );

	    # if source interfaces were specified in the
	    # configuration directive, we have to limit
	    # the possible source interfaces to the
	    # intersection of the interfaces that this 
	    # source address (range) can possible come 
	    # in on, and the interfaces that were specified

	    my @source_interfaces = 
		interfaces_intersection( $vars_ref,
					 \@source_address_interfaces,
					 $source_interfaces_ref );
	    push @$warning_lines_ref,
	    "Source $source can come in on @source_address_interfaces but source interfaces @$source_interfaces_ref were specified, no code generated for this source $for $where"
		if ! @source_interfaces;
	    if ( ${$$vars_ref{constants}}{debug} ) {
	        print "$0: in       generate_source_open_close_lines address_interfaces=<@source_address_interfaces> specified_interfaces=<@$source_interfaces_ref> intersected_interfaces=<@source_interfaces>\n";
	    }

            # if this source/destination is appropriate,
	    # or if this destination is "ANY",
	    # or path optimizations are disabled,
            # add a line to the output commands intermediate
            # file for each protocol/port to close/open it,
	    # and/or add appropriate lines for named
	    # services to open/close them

            if ( ${$$vars_ref{constants}}{nooptpath} ||
		 $destination eq "ANY" ||
		 valid_source_destination_combination 
                 ( $vars_ref,
                   $source, 
                   $destination, 
                   $error_lines_ref,
                   $warning_lines_ref ) ) {

		# one little tricky thing is that if the 
		# destination address is the faked address
		# in an aliased host, the real address 
		# needs to be used instead for the ipchains/
		# iptables/whatever generated rule because
	        # filtering is done in the kernel after
	        # translating the destination address,
	        # this also means re-calculating the
                # list of possible destination interfaces,
		# and we also want to output some comment
		# explanation lines to explain this if
	        # we haven't already output a comment
	        # block explaing this for this destination.
	        # Of course, this is meaningless if the
	        # destination is "ANY".

	        my $alias_faked_to_real_ref = 
		    $$vars_ref{alias_faked_to_real};
		my $translated_destination = $destination;
		my @translated_destination_interfaces = @destination_interfaces;
                if ( $destination ne "ANY" &&
		     exists $$alias_faked_to_real_ref{$destination} ) {

		    $translated_destination = 
			$$alias_faked_to_real_ref{$destination};
		    @translated_destination_interfaces =
			generate_range_to_interfaces
			    ( $vars_ref,
			      $translated_destination,
			      $error_lines_ref,
			      $warning_lines_ref );
		    if ( ! $using_alias_real_notified_flag ) {
			push @$source_lines_ref,
			"  # Since $destination is the pseudo address of an ALIAS,";
			push @$source_lines_ref,
			"  # we use the real address $translated_destination because that is what the";
			    push @$source_lines_ref,
			    "  # packet matching will see after the packet has been remapped.";
			$using_alias_real_notified_flag = 1;
		    }
		}

		# for every source interface, ...

		foreach my $source_interface ( @source_interfaces ) {

		    # for every translated destination interface, ...

		    foreach my $destination_interface ( @translated_destination_interfaces ) {

			# If we're doing output optimizing,
			# and the source interface is not "ANY",
			# and the source interface is not PROTECTED,
			# and the untranslated destination address
			# is not "ANY", and the untranslated 
			# destination address can not come in on 
			# the source interface, this first packet
			# should not be allowed so we won't 
			# generate output lines for it.
			# If any of those conditions don't apply
			# we should generate an output line.

			my $generate_flag = 1;
			if ( ! ${$$vars_ref{constants}}{nooptpath} ) {
#			    print "YEP NOT NOOPTPATH\n";
			    if ( $source_interface ne "ANY" ) {
#				print "YEP SOURCE INTERFACE $source_interface NOT ANY\n";
				my $interfaces_ref =
				    $$vars_ref{interfaces};
				my $source_interface_ref = 
				    $$interfaces_ref{$source_interface};
#				print_hash("", "YEP SOURCE INTERFACE REF", $source_interface_ref);
				if ( ! $$source_interface_ref{protected} ) {
#				    print "YEP SOURCE INTERFACE $source_interface NOT PROTECTED\n";
				    my $source_interface_range = 
					$$source_interface_ref{address};
				    $source_interface_range = 
					"$source_interface_range\/$$source_interface_ref{mask}"
					    if exists $$source_interface_ref{mask} && 
						$$source_interface_ref{mask} ne "";
				    if ( ! range_intersects_range
					 ( $vars_ref,
					   $destination, 
					   $source_interface_range,
					   $error_lines_ref,
					   $warning_lines_ref ) ) {
#					print "YEP NOT RANGE INTERSECTION\n";
					$generate_flag = 0;
				    }
				}
			    }
			}

		        # if we have passed the last decision
		        # to generate some output, ...
		        if ( $generate_flag ) {

			    # if this is an open/close without any 
			    # protocols/ports or defined services,
			    # and without any specified input 
			    # interface or specified output interface,
			    # just use the source and/or destination
			    # with input output interfaces of ANY ANY
			    # and with a protocol/port of ANY/ANY
			    
			    my $protocols_ports_ref = 
				$$directive_ref{destination_protocols_ports};
			    my @protocols = sort keys %$protocols_ports_ref;
			    my $services_ref = $$directive_ref{destination_services};
			    if ( ! @protocols && 
				 ! @$services_ref &&
				 $source_interface eq "ANY" &&
				 $destination_interface eq "ANY" ) {
				push @$source_lines_ref, 
				"$close_open_macro $source $translated_destination ANY ANY ANY ANY";
			    } 
			    
			    # or if there are destination protocols/ports 
			    # and/or destinatoin services listed, insert them
			    
			    else {
				
				# for each protocol listed in the
				# destination protocols/ports hash, ...
				
				foreach my $protocol ( @protocols ) {
				    
				    # for each port listed in the sub-hash
				    
				    my $ports_ref = $$protocols_ports_ref{$protocol};
				    foreach my $port ( sort {$a <=> $b} keys %$ports_ref ) {
					
					# output a source line to open/close
					
					push @$source_lines_ref, 
					"  $close_open_macro $source $translated_destination $source_interface $destination_interface $protocol $port";
					
				    } # end of loop for each port listed
				    
				} # end of loop for each protocol listed
				
				# for each ruleset defined network service listed, ...
				
				foreach my $service ( @$services_ref ) {
				    
				    print "$0: in       generate_source_open_close_lines service=$service action=$ruleset_action\n" 
					if ${$$vars_ref{constants}}{debug};
				
				    # if the appropriate open or close
				    # service portion was defined for this
				    # service, invoke the appropriate 
				    # service macro in the output source
				
			            my $services_ref = $$vars_ref{services};
				    my $service_ref = $$services_ref{$service};
				    if ( exists $$service_ref{$ruleset_action} ) {
				    
					# find out the broadcast addresses for the destination interface

					my $destination_broadcast =
					    generate_interface_to_broadcast
						( $vars_ref,
						  $destination_interface,
						  $error_lines_ref,
						  $warning_lines_ref );
					
					# Since some ruleset defined network services
					# depend on broadcast addresses for incoming
					# packets, if the source interface is marked
					# as "ANY", we want to invoke the ruleset
					# network service open/close macro for each
					# possible incoming packet interface.
					# This will cause more output open/close 
					# commands to be generated than necessary if
					# if the ruleset defined network service
					# doesn't care about broadcast packets, but
					# there is no easy way around that.
					
					my @service_source_interfaces =
					    ( $source_interface eq "ANY" ) ?
					    @{$$vars_ref{interface_names}}
					    :
					    ( $source_interface );
					
					# for each source interface, ...
					
					foreach my $service_source_interface ( @service_source_interfaces ) {
					    
					    # find out the broadcast addresses for the source interface
					    
					    my $source_broadcast =
						generate_interface_to_broadcast
						    ( $vars_ref,
						      $service_source_interface,
						      $error_lines_ref,
						      $warning_lines_ref );
					    
					    # finally, invoke the service macro
					    
					    push @$source_lines_ref,
					    "  %$$service_ref{$ruleset_action}% $source $translated_destination $service_source_interface $destination_interface $source_broadcast $destination_broadcast";
					    
					    
					} # end of loop for each service source interface
					
				    } # end if this service portion ( "open" or "close" ) existed
				
			        } # end of loop for each service listed
			   
			    } # end if not an ANY ANY ANY ANY

		        } # end if the last decision was to generate output

		    } # end for each destination interface

		} # end for each source interface

            } # end if the source/destination is appropriate

        } # end of loop for each source address
    
    } # end of loop for each destination address

    print "$0: exiting  generate_source_open_close_lines\n" 
        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================

# function to generate code for open/close/alias/literal/forward directives
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_standard_directives {
    my ( $vars_ref,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;
    print "$0: entering generate_source_standard_directives\n"
	if ${$$vars_ref{constants}}{debug};

    # now add open and close ports/services macros corresponding
    # to the directives that were parsed from the configuration file.
    # also add pseudo interface creation and route addition commands
    # caused by alias directives. plus translation commands that
    # are caused by the forward directives. plus code for forwarding
    # packets caused to forward directives.
    # we also want to intersperse any comment blocks that preceded
    # the directives we are using here to generate the opens/closes.
    # scan through the stored configturation file directives.
    # another tricky thing is that an open/close with a destination
    # of an aliased hidden machine can specify either the real or
    # the faked addresses but end up doing the same thing.
    my $pseudo_interface_index = 1;
    # loop through the directives
    foreach my $directive_ref ( @{$$vars_ref{config}} ) {
	# initialize generated source lines for this directive
	my @source_lines = ();
        # figure out which type of directive this one is
        my $directive_type = $$directive_ref{directive_type};
        # or if this is an ALIAS which specified ports/services ...
        if ( $directive_type eq "alias" ) {
	    generate_source_alias_directive( $vars_ref,
					     $directive_ref,
					     $pseudo_interface_index,
					     \@source_lines,
					     $error_lines_ref,
					     $warning_lines_ref );
	    # update the index of the pseudo interface to be created
	    $pseudo_interface_index++;
	} # end if this is an ALIAS directive
        # if this is a CLOSE or OPEN directive ...
        elsif ( $directive_type eq "close" ||
             $directive_type eq "open" ) {
	    generate_source_open_close_directive( $vars_ref,
						  $directive_ref,
						  \@source_lines,
						  $error_lines_ref,
						  $warning_lines_ref );
        } # end if this is a CLOSE or OPEN directive
        # if this is a FORWARD directive ...
        elsif ( $directive_type eq "forward" ) {
	    generate_source_forward_directive( $vars_ref,
					       $directive_ref,
					       \@source_lines,
					       $error_lines_ref,
					       $warning_lines_ref );
        } # end if this is a FORWARD directive
        # if this is a LITERAL block ...
        elsif ( $directive_type eq "literal_block" ) {
	    generate_source_literal_block( $vars_ref,
					   $directive_ref,
					   \@source_lines,
					   $error_lines_ref,
					   $warning_lines_ref );
        } # end if this is a LITERAL block
        # or if this is a TRUSTED/UNTRUSTED/FILTERED/UNFILTERED
	# which specified one or more address (range)(s) ...
        elsif ( ( $directive_type eq "filtered"   ||
		  $directive_type eq "trusted"    ||
		  $directive_type eq "unfiltered" ||
		  $directive_type eq "untrusted" )
		&& exists $$directive_ref{source_addresses} ) {
	    generate_source_trust_address_directive( $vars_ref,
						     $directive_ref,
						     \@source_lines,
						     $error_lines_ref,
						     $warning_lines_ref );
	} # end if TRUSTED/UNTRUSTED/FILTERED/UNFILTERED directive
        # or if this is a directive that had a syntax error, ...
        elsif ( $directive_type eq "syntax_error" ) {
	    generate_source_syntax_error_directive( $vars_ref,
						    $directive_ref,
						    \@source_lines,
						    $error_lines_ref,
						    $warning_lines_ref );
	} # end if this is a syntax error directive
	# add any generated warning/error lines
	# to the total generated source lines list
	if ( exists $$directive_ref{error_lines} 
	     && @{$$directive_ref{error_lines}} ) {
	    push @source_lines, "  #";
	    push @source_lines, "  # $$directive_ref{config_where} error(s):";
	    foreach my $error_line ( @{$$directive_ref{error_lines}} ) {
		push @source_lines, "  #   $error_line";
	    }
	}
	if ( exists $$directive_ref{warning_lines}
	     && @{$$directive_ref{warning_lines}} ) {
	    push @source_lines, "  #";
	    push @source_lines, "  # $$directive_ref{config_where} warning(s):";
	    foreach my $warning_line ( @{$$directive_ref{warning_lines}} ) {
		push @source_lines, "  #   $warning_line";
	    }
	}
	# if some output was generated, insert a blank line and the
	# generated output to the total generated source lines list
	if ( @source_lines ) {
	    push @source_lines, "";
	    push @$source_lines_ref, @source_lines;
	}
    } # end looping through the directives
    
    print "$0: exiting  generate_source_standard_directives\n" 
        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================

# function to generate the source code for a syntax error directive
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to directive being used to generate code
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_syntax_error_directive {
    my ( $vars_ref,
	 $directive_ref,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;
    print "$0: entering generate_source_syntax_error_directive\n"
	if ${$$vars_ref{constants}}{debug};

    # mark this directive as being processed
    $$directive_ref{processed} = 1;
    print_hash("", "generating output from directive", $directive_ref)
        if ${$$vars_ref{constants}}{debug};

    # output the original config file source line as a comment
    push @$source_lines_ref,
    "  # $$directive_ref{config_line_number}\t$$directive_ref{config_line}";
    push @$source_lines_ref, "  #";

    # output a syntax error comment line
    push @$source_lines_ref, "  #";
    push @$source_lines_ref, 
    "  # $$directive_ref{config_where} was not accepted because of a syntax error";

    print "$0: exiting  generate_source_syntax_error_directive\n" 
        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================

# function to generate the source code for the portions of a
# trusted/untrusted/filtered/unfiltered directives that specify
# source address (range)(s). This does not do anything about the
# portions of trusted/untrusted/filtered/unfiltered directives that 
# specify interface(s).
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to directive being used to generate code
#   reference to source lines list to add generated lines to
#   reference to error lines array
#   reference to warning lines array

sub generate_source_trust_address_directive {
    my ( $vars_ref,
	 $directive_ref,
	 $source_lines_ref,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;
    print "$0: entering generate_source_trust_address_directive\n"
	if ${$$vars_ref{constants}}{debug};

    # figure out which type of directive this one is
    my $directive_type = $$directive_ref{directive_type};
    # mark this directive as being processed
    $$directive_ref{processed} = 1;
    print_hash("", "generating output from directive", $directive_ref)
        if ${$$vars_ref{constants}}{debug};
    # pick the macro to invoke to open/close this protocol/port
    # for unfiltered and trusted directives we're closing ports,
    # for filtered and untrusted directives we're closing ports
    my $close_open_macro = ( $directive_type eq "untrusted" ||
                             $directive_type eq "filtered" ) ?
                             "%close_protocol_port%" : 
                             "%open_protocol_port%";
    # put a line in the output showing the original configuration
    # file line that caused the following output to be generated
    push @$source_lines_ref,
    "  # $$directive_ref{config_line_number}\t$$directive_ref{config_line}";
    push @$source_lines_ref, "  #";
    # there are always source addresses if we're here
    my @sources = @{$$directive_ref{source_addresses}};
    # figure out the destination addresses, it's either
    # specified to be a specific address(range), or no 
    # destination address was specified which means that 
    # the destination is this computer so in that case
    # we set the destinations to all the filtered
    # interface addresses since the unfiltered interfaces
    # accept everything going to this machine or to ALL
    # if we are in --nooptpath mode
    my @destinations;
    if ( exists $$directive_ref{destination_addresses} &&
	 scalar @{$$directive_ref{destination_addresses}} ) {
        @destinations = @{$$directive_ref{destination_addresses}};
    } else {
	@destinations = ( ${$$vars_ref{constants}}{nooptpath} ) ?
	    ( "ANY" ) : @{$$vars_ref{interface_addresses}};
    }
    # for each source address, ...
    foreach my $source ( @sources ) {
        # for each destination address, ...
        foreach my $destination ( @destinations ) {
            # insert a line in the intermediate source to close or open
#            print "$0: inserting internmediate source line for source $source\n" 
#                if ${$$vars_ref{constants}}{debug};
            push @$source_lines_ref, 
            "  $close_open_macro $source $destination ANY ANY ANY ANY";
        } # end of loop for each destination address
    } # end of loop for each source address
    
    print "$0: exiting  generate_source_trust_address_directive\n" 
        if ${$$vars_ref{constants}}{debug};
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"
    
1;


