# Pfilter/Protocols.pl
#
# library functions to handle pfilter configuration files

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# functions in this file:

# PARSE /etc/protocols FILE and find named services/protocols/ports
# sub parse_etc_protocols( $vars_ref, $protocols_path )

#=======================================================================
#
# PARSE /etc/protocols FILE and find named services/protocols/ports
#
# Passed input parameters are:
#   reference to hash of useful values:
#   protocols file path name
# Returned parameters are:
#   reference to named services/protocols/ports hash

sub parse_etc_protocols
{
    my( $vars_ref, $protocols_path ) = @_;
#    print "$0: in parse_etc_protocols\n" if $debug;

    my %protocols;
    open(PROTOCOLS, $protocols_path) ||
	die "$0: fatal error, cannot open protocols file $protocols_path for reading\n";
    while (<PROTOCOLS>) {
	chomp;                      # get rid of trailing newline characters
	s/\s*#.*$//;                # get rid of comments
	next unless /\w/;           # skip lines without any protocols in them
	my @names = split;          # get all names on line
	foreach my $protocol ( @names ) {
	    next if $protocol =~ /^\d+$/;
	    my $name = lc $protocol;
	    $protocols{$name} = 1;
	}
    }
    close(PROTOCOLS);

    print "$0: cannot find any protocols from $protocols_path file\n"
        unless %protocols;
    return \%protocols;
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;

