# Pfilter/Services.pl
#
# library functions to handle pfilter configuration files

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# functions in this file

# function to PARSE /etc/services FILE and find named services/protocols/ports
# sub parse_etc_services( $vars_ref, $services_path )

#=======================================================================
#
# function to PARSE /etc/services FILE and find named  services/protocols/ports
#
# Passed input parameters are:
#   reference to hash of useful values:
#   services file path name
# Returned parameters are:
#   reference to named services/protocols/ports hash

sub parse_etc_services
{
    my( $vars_ref, $services_path ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my $verbose = ${$$vars_ref{constants}}{verbose}; # save time
    my %named_services_protocols_ports;
#    print "$0: in parse_etc_services\n" if $debug;

    open(SERVICES, $services_path) ||
	die "$0: fatal error, cannot open services file $services_path for reading\n";
    while (<SERVICES>) {		# for every line in the service file
	chomp;				# get rid of trailing new-line chars
	s/\s*#.*$//;			# get rid of comments
	next unless /\w/;		# skip lines without any services in them
	$_ = lc $_;			# named services are case independant
	my ($service_name, $port_protocol, @aliases) = split;
	unless ( defined $port_protocol ) {	# has to be at least two fields
	    print "$0: too few fields in $services_path line $.\n"
		if $verbose;
	    next;
	}
	my ($port, $protocol) = split( /\//, $port_protocol );	# find port and protocol
	unless ( defined $protocol ) {	# port/protocol has to be correct
	    print "$0: too few fields in $services_path line $.\n"
		if $verbose;
	    next;
	}
	unless ( $port =~ /^\d+$/ ) {
	    print "$0: non-numeric port $port in $services_path line $.\n"
		if $verbose;
	    next;
	}
	unless ( ${$$vars_ref{protocols}}{$protocol} ) {
	    print "$0: unknown protocol $protocol in $services_path line $.\n"
		if $verbose;
	    next;
	}
        foreach my $service ( $service_name, @aliases ) {
	    my %ports;
	    $ports{$port} = 1;
	    my %protocols_ports;
	    $protocols_ports{$protocol} = \%ports;
	    if ( exists $named_services_protocols_ports{$service} ) {
		protocols_ports_add_protocols_ports
		    ( \%vars, $named_services_protocols_ports{$service},
		      \%protocols_ports );
	    } else {
		$named_services_protocols_ports{$service} = \%protocols_ports;
	    }
	}
    }
    close(SERVICES);

    print "$0: cannot find any service protocols/ports from $services_path file\n"
        unless %named_services_protocols_ports;
    return \%named_services_protocols_ports;
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;

